// File     CNamedSemaphore.cpp
// Version  1.0
// Author   Jens-Peter Akelbein
// Comment  Softwareentwicklung fuer Embedded Systeme - Exercise 3

#include <iostream>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include "CNamedSemaphore.h"
#include <unistd.h>
#include <iostream>

using namespace std;

// construct a new named semaphore
// @name - name of the semaphore being created in /run/shm
// @value - initial value of the semaphore
CNamedSemaphore::CNamedSemaphore(const char *name, int value) {


    semaphore = sem_open(name, O_CREAT|O_EXCL, 0644, value);
    if (semaphore == SEM_FAILED)
        exitproc("Error opening", errno);

    remember_my_creator = getpid();
    remember_my_name= name;

}

// deconstruct the semaphore
// question:
// How do we delete the semaphore by the last process only?
CNamedSemaphore::~CNamedSemaphore() {

    cout << "Destructor for " << remember_my_name << " Called by: ";

    if (sem_close(semaphore) == -1)
        exitproc("Error closing", errno);

    if (getpid() == remember_my_creator) {
        cout << " PARENT" << endl;
        if (sem_unlink(remember_my_name) == -1)
            exitproc("Error unlinking", errno);

    } else {
        cout << " CHILD" << endl;
    }

}


void CNamedSemaphore::increment(void) {
    if (sem_post(semaphore) == -1)
        exitproc("Error INCREMENTING", errno);

}


void CNamedSemaphore::decrement(void) {

    if (sem_wait(semaphore) == -1)
        exitproc("Error DECREMENTING", errno);
}


int CNamedSemaphore::value(void) {
    int val;
    if (sem_getvalue(semaphore, &val) == -1)
        exitproc("Error getting value", errno);

    return val;
}


// helper function to display errors and terminate our
// process as a very simple error handling mechanism
void CNamedSemaphore::exitproc(const char *text, int err) {

    cout << text << ": ";
    switch (err) {
        case EACCES:
            cerr << "EACCES";
            break;
        case EEXIST:
            cerr << "EEXIST";
            break;
        case EINVAL:
            cerr << "EINVAL";
            break;
        case EMFILE:
            cerr << "EMFILE";
            break;
        case ENAMETOOLONG:
            cerr << "TOOLNG";
            break;
        case ENOENT:
            cerr << "ENOENT";
            break;
        case ENOMEM:
            cerr << "ENOMEM";
            break;
        case EOVERFLOW:
            cerr << "EOVERFLOW";
            break;
        default:
            cerr << "UNKNWN";
    }
    cout << endl;
    exit(1);
}
